@@ -3,12 +3,14 @@
 
 #include <QtTest/qtest.h>
 #include <QtTest/qsignalspy.h>
+#include <QtQuickTest/quicktest.h>
 
 #include <QtCore/qoperatingsystemversion.h>
 #include <QtGui/qpa/qwindowsysteminterface.h>
 #include <QtGui/qpa/qplatformintegration.h>
 #include <QtGui/private/qguiapplication_p.h>
 #include <QtQuick/qquickview.h>
+#include <QtQuick/private/qquickmousearea_p.h>
 #include <QtQuick/private/qquickpalette_p.h>
 #include <QtQuickTestUtils/private/qmlutils_p.h>
 #include <QtQuickTestUtils/private/viewtestutils_p.h>
@@ -25,6 +27,10 @@
 #include <QtQuickTemplates2/private/qquickstackview_p.h>
 #include <QtQuickTemplates2/private/qquickpopup_p_p.h>
 #include <QtQuickTemplates2/private/qquicktooltip_p.h>
+#include <QtQuickTemplates2/private/qquickdrawer_p.h>
+#include <QtQuick/private/qquicklistview_p.h>
+#include <QtQuick/private/qquicktextedit_p.h>
+#include <QtQuick/private/qquickdroparea_p.h>
 #include <QtQuickControlsTestUtils/private/controlstestutils_p.h>
 #include <QtQuickControlsTestUtils/private/qtest_quickcontrols_p.h>
 
@@ -66,6 +72,7 @@ private slots:
     void parentDestroyed();
     void nested();
     void nestedWheel();
+    void nestedWheelWithOverlayParent();
     void modelessOnModalOnModeless();
     void grabber();
     void cursorShape();
@@ -92,6 +99,12 @@ private slots:
     void relativeZOrder();
     void mirroredCombobox();
     void rotatedCombobox();
+    void focusMultiplePopup();
+    void contentChildrenChange();
+    void doubleClickInMouseArea();
+    void fadeDimmer_data();
+    void fadeDimmer();
+    void noDimmer();
 
 private:
     static bool hasWindowActivation();
@@ -217,6 +230,10 @@ void tst_QQuickPopup::overlay_data()
 
 void tst_QQuickPopup::overlay()
 {
+#ifdef Q_OS_ANDROID
+    QSKIP("Test crashes. See QTBUG-118532");
+#endif
+
     QFETCH(QString, source);
     QFETCH(bool, modal);
     QFETCH(bool, dim);
@@ -977,18 +994,19 @@ void tst_QQuickPopup::hover()
     QVERIFY(openedSpy.size() == 1 || openedSpy.wait());
     QTRY_VERIFY(popup->width() > 10); // somehow this can take a short time with macOS style
 
-    // hover the parent button outside the popup
-    QTest::mouseMove(window, QPoint(window->width() - 1, window->height() - 1));
+    // Hover the parent button outside the popup. It has 10 pixel anchor margins around the window.
+    PointLerper pointLerper(window);
+    pointLerper.move(15, 15);
     QCOMPARE(parentButton->isHovered(), !modal);
     QVERIFY(!childButton->isHovered());
 
-    // hover the popup background
-    QTest::mouseMove(window, QPoint(1, 1));
+    // Hover the popup background. Its top-left is 10 pixels in from its parent.
+    pointLerper.move(25, 25);
     QVERIFY(!parentButton->isHovered());
     QVERIFY(!childButton->isHovered());
 
-    // hover the child button in a popup
-    QTest::mouseMove(window, QPoint(popup->x() + popup->width() / 2, popup->y() + popup->height() / 2));
+    // Hover the child button in a popup.
+    pointLerper.move(mapCenterToWindow(childButton));
     QVERIFY(!parentButton->isHovered());
     QVERIFY(childButton->isHovered());
 
@@ -1183,6 +1201,41 @@ void tst_QQuickPopup::nestedWheel()
     QTRY_COMPARE_GT(vbar->property("position").toDouble(), startPosition);
 }
 
+void tst_QQuickPopup::nestedWheelWithOverlayParent()
+{
+    QQuickControlsApplicationHelper helper(this, QStringLiteral("nested-wheel-overlay-parent.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowExposed(window));
+
+    auto *drawer= window->property("_drawer").value<QQuickDrawer *>();
+    QVERIFY(drawer);
+
+    auto *dropArea = window->property("_dropArea").value<QQuickDropArea *>();
+    QVERIFY(dropArea);
+
+    drawer->open();
+    QCOMPARE(drawer->isVisible(), true);
+    QTRY_COMPARE(drawer->isOpened(), true);
+
+    QQuickListView *listView = window->property("_listView").value<QQuickListView *>();
+    QTRY_VERIFY(listView != nullptr);
+    QQuickItem *contentItem = listView->contentItem();
+    QTRY_VERIFY(contentItem != nullptr);
+
+    // Check parent is set as overlay
+    QTRY_COMPARE(dropArea->parentItem(), QQuickOverlay::overlay(window));
+    // Consider the center point of the control as event position to trigger wheel event
+    QVERIFY(sendWheelEvent(listView, -15));
+
+    if (QQuickTest::qIsPolishScheduled(listView))
+        QVERIFY(QQuickTest::qWaitForPolish(listView));
+
+    // Wheel over the list view, verify that it scrolls
+    QTRY_COMPARE(listView->contentY(), 72.);
+}
+
 void tst_QQuickPopup::modelessOnModalOnModeless()
 {
     QQuickControlsApplicationHelper helper(this, QStringLiteral("modelessOnModalOnModeless.qml"));
@@ -2149,6 +2202,152 @@ void tst_QQuickPopup::rotatedCombobox()
     }
 }
 
+void tst_QQuickPopup::focusMultiplePopup()
+{
+    QQuickApplicationHelper helper(this, "multiplepopup.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowExposed(window));
+
+    auto *rootItem = window->findChild<QQuickItem *>("rootItem");
+    QTRY_VERIFY(rootItem->hasFocus());
+
+    auto *buttonPopup = window->findChild<QQuickPopup *>("popup1");
+    buttonPopup->open();
+    QTRY_VERIFY(buttonPopup->isOpened());
+    QVERIFY(!rootItem->hasFocus());
+    QVERIFY(buttonPopup->hasFocus());
+
+    auto *textEditPopup = window->findChild<QQuickPopup *>("popup2");
+    textEditPopup->open();
+    QTRY_VERIFY(textEditPopup->isOpened());
+    QVERIFY(textEditPopup->hasFocus());
+
+    auto *drawerPopup = window->findChild<QQuickPopup *>("popup3");
+    drawerPopup->open();
+    QTRY_VERIFY(drawerPopup->isVisible());
+    QVERIFY(drawerPopup->hasFocus());
+
+    auto *drawer = window->findChild<QQuickDrawer *>("drawer");
+    drawer->close();
+    QTRY_VERIFY(!drawer->isVisible());
+    drawerPopup->close();
+    QTRY_VERIFY(!drawerPopup->isVisible());
+    QVERIFY(textEditPopup->hasFocus());
+
+    textEditPopup->close();
+    QTRY_VERIFY(!textEditPopup->isVisible());
+    QVERIFY(buttonPopup->hasFocus());
+
+    buttonPopup->close();
+    QTRY_VERIFY(!buttonPopup->isVisible());
+
+    QVERIFY(rootItem->hasFocus());
+}
+
+void tst_QQuickPopup::contentChildrenChange()
+{
+    QQmlEngine engine;
+    QQmlComponent comp(&engine);
+    comp.loadFromModule("QtQuick.Controls", "Popup");
+    std::unique_ptr<QObject> root {comp.create()};
+    QVERIFY(root);
+    QQuickPopup *popup = qobject_cast<QQuickPopup *>(root.get());
+    QVERIFY(popup);
+    QSignalSpy spy(popup, &QQuickPopup::contentChildrenChanged);
+    auto contentItem = std::make_unique<QQuickItem>();
+    popup->setContentItem(contentItem.get());
+    QCOMPARE(spy.count(), 1);
+    auto newChild = std::make_unique<QQuickItem>();
+    QQmlProperty contentItemChildren(contentItem.get());
+    contentItemChildren.write(QVariant::fromValue(newChild.get()));
+    QCOMPARE(spy.count(), 2);
+}
+
+void tst_QQuickPopup::doubleClickInMouseArea()
+{
+#ifdef Q_OS_ANDROID
+    QSKIP("Test crashes. See QTBUG-118532");
+#endif
+
+    QQuickView window;
+    QVERIFY(QQuickTest::showView(window, testFileUrl("doubleClickInMouseArea.qml")));
+
+    auto *ma = window.rootObject()->findChild<QQuickMouseArea *>();
+    QVERIFY(ma);
+    QSignalSpy doubleClickSpy(ma, &QQuickMouseArea::doubleClicked);
+    QSignalSpy longPressSpy(ma, &QQuickMouseArea::pressAndHold);
+    QPoint p = ma->mapToScene(ma->boundingRect().center()).toPoint();
+
+    // check with normal double click
+    QTest::mouseDClick(&window, Qt::LeftButton, Qt::NoModifier, p);
+    QCOMPARE(doubleClickSpy.count(), 1);
+
+    // wait enough time for a wrong long press to happen
+    QTest::qWait(QGuiApplication::styleHints()->mousePressAndHoldInterval() + 10);
+    QCOMPARE(longPressSpy.count(), 0);
+}
+
+void tst_QQuickPopup::fadeDimmer_data()
+{
+    QTest::addColumn<bool>("modality");
+
+    QTest::addRow("modal") << true;
+    QTest::addRow("modeless") << true;
+}
+
+void tst_QQuickPopup::fadeDimmer()
+{
+    QFETCH(const bool, modality);
+    QQuickApplicationHelper helper(this, "fadeDimmer.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowExposed(window));
+
+    auto *popup = window->contentItem()->findChild<QQuickPopup *>();
+    QVERIFY(popup);
+
+    popup->setModal(modality);
+    popup->open();
+    auto dimmer = QQuickPopupPrivate::get(popup)->dimmer;
+    QVERIFY(dimmer);
+    int opacityChangeCount = 0;
+    connect(dimmer, &QQuickItem::opacityChanged, this, [&opacityChangeCount]{
+        ++opacityChangeCount;
+    });
+    QTRY_VERIFY(popup->isOpened());
+    QTRY_COMPARE(dimmer->opacity(), popup->property("dimmerOpacity").toDouble());
+    QCOMPARE_GT(opacityChangeCount, 2);
+
+    opacityChangeCount = 0;
+    popup->setVisible(false);
+    QTRY_VERIFY(!popup->isVisible());
+    QCOMPARE_GT(opacityChangeCount, 2);
+}
+
+void tst_QQuickPopup::noDimmer()
+{
+    QQuickApplicationHelper helper(this, "noDimmer.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowExposed(window));
+
+    auto *drawer = window->contentItem()->findChild<QQuickDrawer *>();
+    QVERIFY(drawer);
+
+    drawer->open();
+    auto dimmer = QQuickPopupPrivate::get(drawer)->dimmer;
+    QVERIFY(dimmer);
+    // this must not crash
+    QTRY_VERIFY(!drawer->isModal());
+}
+
 QTEST_QUICKCONTROLS_MAIN(tst_QQuickPopup)
 
 #include "tst_qquickpopup.moc"
