@@ -89,8 +89,19 @@ private slots:
     void setSelectedFile();
     void selectNewFileViaTextField_data();
     void selectNewFileViaTextField();
+    void selectExistingFileShouldWarnUserWhenFileModeEqualsSaveFile();
+    void fileNameTextFieldOnlyChangesWhenSelectingFiles();
 
 private:
+    enum DelegateOrderPolicy
+    {
+        ShowDirectoriesFirst,
+        ShowFilesFirst
+    };
+
+    QStringList tempDirExpectedVisibleFiles(DelegateOrderPolicy order) const;
+    QStringList tempSubDirExpectedVisibleFiles(DelegateOrderPolicy order) const;
+
     QTemporaryDir tempDir;
     QScopedPointer<QFile> tempFile1;
     QScopedPointer<QFile> tempFile2;
@@ -110,6 +121,20 @@ private:
     const QKeySequence editPathKeySequence = QKeySequence(Qt::CTRL | Qt::Key_L);
 };
 
+QStringList tst_QQuickFileDialogImpl::tempDirExpectedVisibleFiles(DelegateOrderPolicy order) const
+{
+    return order == ShowDirectoriesFirst
+        ? QStringList { tempSubDir.path(), tempFile1->fileName(), tempFile2->fileName() }
+        : QStringList { tempFile1->fileName(), tempFile2->fileName(), tempSubDir.path() };
+}
+
+QStringList tst_QQuickFileDialogImpl::tempSubDirExpectedVisibleFiles(DelegateOrderPolicy order) const
+{
+    return order == ShowDirectoriesFirst
+        ? QStringList { tempSubSubDir.path(), tempSubFile1->fileName(), tempSubFile2->fileName() }
+        : QStringList { tempSubFile1->fileName(), tempSubFile2->fileName(), tempSubSubDir.path() };
+}
+
 // We don't want to fail on warnings until QTBUG-98964 is fixed,
 // as we deliberately prevent deferred execution in some of the tests here,
 // which causes warnings.
@@ -135,13 +160,13 @@ void tst_QQuickFileDialogImpl::initTestCase()
     /*
         Create a couple of files within the temporary directory. The structure is:
 
-        [temp directory]
-            ├── sub-dir
-            │   ├── sub-sub-dir
-            │   ├── sub-file1.txt
-            │   └── sub-file2.txt
-            ├── file1.txt
-            └── file2.txt
+        [temp directory] (tempDir)
+            ├── sub-dir (tempSubDir)
+            │   ├── sub-sub-dir (tempSubSubDir)
+            │   ├── sub-file1.txt (tempSubFile1)
+            │   └── sub-file2.txt (tempSubFile2)
+            ├── file1.txt (tempFile1)
+            └── file2.txt (tempFile2)
     */
     tempSubDir = QDir(tempDir.path());
     QVERIFY2(tempSubDir.mkdir("sub-dir"), qPrintable(QString::fromLatin1(
@@ -232,8 +257,8 @@ void tst_QQuickFileDialogImpl::cleanupTestCase()
     QTRY_VERIFY(findViewDelegateItem(dialogHelper.fileDialogListView, expectedCurrentIndex, fileDelegate)); \
     QVERIFY2(fileDelegate->hasActiveFocus(), qPrintable(QString::fromLatin1( \
         "Expected delegate at index %1 to have focus, but %2 has it") \
-            .arg(QString::number(expectedCurrentIndex)) \
-            .arg(QDebug::toString(dialogHelper.window()->activeFocusItem())))); \
+            .arg(QString::number(expectedCurrentIndex), \
+                QDebug::toString(dialogHelper.window()->activeFocusItem())))); \
     QVERIFY(fileDelegate->isHighlighted()); \
 }
 
@@ -418,8 +443,8 @@ void tst_QQuickFileDialogImpl::bindCurrentFolder_data()
     const auto tempSubDirUrl = QUrl::fromLocalFile(tempSubDir.path());
     const auto tempSubFile1Url = QUrl::fromLocalFile(tempSubFile1->fileName());
 
-    const QStringList currentDirFiles = { tempSubDir.path(), tempFile1->fileName(), tempFile2->fileName() };
-    const QStringList tempSubDirFiles = { tempSubSubDir.path(), tempSubFile1->fileName(), tempSubFile2->fileName() };
+    const QStringList currentDirFiles = tempDirExpectedVisibleFiles(ShowDirectoriesFirst);
+    const QStringList tempSubDirFiles = tempSubDirExpectedVisibleFiles(ShowDirectoriesFirst);
 
     // Setting the folder to "sub-dir" should result in "sub-file1.txt" and "sub-file2.txt" being visible.
     QTest::addRow("sub-dir") << tempSubDirUrl << tempSubDirUrl << tempSubDirFiles;
@@ -505,9 +530,7 @@ void tst_QQuickFileDialogImpl::changeFolderViaDoubleClick()
     QTRY_VERIFY(findViewDelegateItem(dialogHelper.fileDialogListView, subDirIndex, subDirDelegate));
     COMPARE_URL(subDirDelegate->file(), QUrl::fromLocalFile(tempSubDir.path()));
     QVERIFY(doubleClickButton(subDirDelegate));
-    const QStringList expectedVisibleFiles = showDirsFirst
-        ? QStringList { tempSubSubDir.path(), tempSubFile1->fileName(), tempSubFile2->fileName() }
-        : QStringList { tempSubFile1->fileName(), tempSubFile2->fileName(), tempSubSubDir.path() };
+    const QStringList expectedVisibleFiles = tempSubDirExpectedVisibleFiles(showDirsFirst ? ShowDirectoriesFirst : ShowFilesFirst);
     QString failureMessage;
     QTRY_VERIFY2(verifyFileDialogDelegates(dialogHelper.fileDialogListView, expectedVisibleFiles, failureMessage), qPrintable(failureMessage));
     // The first file in the directory should now be selected.
@@ -544,6 +567,9 @@ void tst_QQuickFileDialogImpl::chooseFolderViaTextEdit()
 
     // Hit enter to accept.
     QTest::keyClick(dialogHelper.window(), Qt::Key_Return);
+    QString failureMessage;
+    QTRY_VERIFY2(verifyFileDialogDelegates(dialogHelper.fileDialogListView,
+        tempSubDirExpectedVisibleFiles(ShowDirectoriesFirst), failureMessage), qPrintable(failureMessage));
     // The first file in the directory should be selected, which is "sub-sub-dir".
     // Note that the TextEdit will still have focus, so we can't use VERIFY_FILE_SELECTED_AND_FOCUSED.
     VERIFY_FILE_SELECTED(QUrl::fromLocalFile(tempSubDir.path()), QUrl::fromLocalFile(tempSubSubDir.path()));
@@ -566,9 +592,9 @@ void tst_QQuickFileDialogImpl::chooseFolderViaEnter()
 
     // Select the delegate by pressing enter.
     QTest::keyClick(dialogHelper.window(), Qt::Key_Return);
-    const QStringList expectedVisibleFiles = { tempSubSubDir.path(), tempSubFile1->fileName(), tempSubFile2->fileName() };
     QString failureMessage;
-    QTRY_VERIFY2(verifyFileDialogDelegates(dialogHelper.fileDialogListView, expectedVisibleFiles, failureMessage), qPrintable(failureMessage));
+    QTRY_VERIFY2(verifyFileDialogDelegates(dialogHelper.fileDialogListView,
+        tempSubDirExpectedVisibleFiles(ShowDirectoriesFirst), failureMessage), qPrintable(failureMessage));
     // The first file in the new directory should be selected, which is "sub-sub-dir".
     VERIFY_FILE_SELECTED_AND_FOCUSED(QUrl::fromLocalFile(tempSubDir.path()), QUrl::fromLocalFile(tempSubSubDir.path()), 0);
     // Since we only chose a folder, the dialog should still be open.
@@ -628,6 +654,9 @@ void tst_QQuickFileDialogImpl::chooseFileAndThenFolderViaTextEdit()
 
     // Hit enter to accept.
     QTest::keyClick(dialogHelper.window(), Qt::Key_Return);
+    QString failureMessage;
+    QTRY_VERIFY2(verifyFileDialogDelegates(dialogHelper.fileDialogListView,
+        tempSubDirExpectedVisibleFiles(ShowDirectoriesFirst), failureMessage), qPrintable(failureMessage));
     // The first file in the directory should be selected, which is "sub-sub-dir".
     // Note that the TextEdit will still have focus, so we can't use VERIFY_FILE_SELECTED_AND_FOCUSED.
     VERIFY_FILE_SELECTED(QUrl::fromLocalFile(tempSubDir.path()), QUrl::fromLocalFile(tempSubSubDir.path()));
@@ -727,6 +756,9 @@ void tst_QQuickFileDialogImpl::goUp()
         QVERIFY(QQuickTest::qWaitForPolish(barListView));
     QVERIFY(clickButton(breadcrumbBar->upButton()));
     // The previous directory that we were in should now be selected (matches e.g. Windows and Ubuntu).
+    QString failureMessage;
+    QTRY_VERIFY2(verifyFileDialogDelegates(dialogHelper.fileDialogListView,
+        tempDirExpectedVisibleFiles(showDirsFirst ? ShowDirectoriesFirst : ShowFilesFirst), failureMessage), qPrintable(failureMessage));
     int expectedCurrentIndex = showDirsFirst ? 0 : 2;
     VERIFY_FILE_SELECTED_AND_FOCUSED(QUrl::fromLocalFile(tempDir.path()), QUrl::fromLocalFile(tempSubDir.path()), expectedCurrentIndex);
 
@@ -878,7 +910,7 @@ void tst_QQuickFileDialogImpl::bindNameFilters()
     // Only "sub-dir", "text1.txt" and "text2.txt" should be visible, since *.txt is the first filter.
     QString failureMessage;
     QTRY_VERIFY2(verifyFileDialogDelegates(dialogHelper.fileDialogListView,
-        { tempSubDir.path(), tempFile1->fileName(), tempFile2->fileName() }, failureMessage), qPrintable(failureMessage));
+        tempDirExpectedVisibleFiles(ShowDirectoriesFirst), failureMessage), qPrintable(failureMessage));
 }
 
 void tst_QQuickFileDialogImpl::changeNameFilters()
@@ -929,7 +961,7 @@ void tst_QQuickFileDialogImpl::changeNameFilters()
     // Only "sub-dir", "text1.txt" and "text2.txt" should be visible, since *.txt is the first filter.
     QString failureMessage;
     QTRY_VERIFY2(verifyFileDialogDelegates(dialogHelper.fileDialogListView,
-        { tempSubDir.path(), tempFile1->fileName(), tempFile2->fileName() }, failureMessage), qPrintable(failureMessage));
+        tempDirExpectedVisibleFiles(ShowDirectoriesFirst), failureMessage), qPrintable(failureMessage));
 
     // Open the ComboBox's popup.
     const QQuickComboBox *comboBox = dialogHelper.quickDialog->findChild<QQuickComboBox*>();
@@ -964,7 +996,7 @@ void tst_QQuickFileDialogImpl::changeNameFilters()
     }
     QTRY_VERIFY(!comboBox->popup()->isVisible());
     QTRY_VERIFY2(verifyFileDialogDelegates(dialogHelper.fileDialogListView,
-        { tempSubDir.path(), tempFile1->fileName(), tempFile2->fileName() }, failureMessage), qPrintable(failureMessage));
+        tempDirExpectedVisibleFiles(ShowDirectoriesFirst), failureMessage), qPrintable(failureMessage));
 }
 
 void tst_QQuickFileDialogImpl::changeNameFiltersAfterChangingFolder()
@@ -976,7 +1008,7 @@ void tst_QQuickFileDialogImpl::changeNameFiltersAfterChangingFolder()
     // Go into the "sub-dir" folder.
     QString failureMessage;
     QTRY_VERIFY2(verifyFileDialogDelegates(dialogHelper.fileDialogListView,
-        { tempSubDir.path(), tempFile1->fileName(), tempFile2->fileName() }, failureMessage), qPrintable(failureMessage));
+        tempDirExpectedVisibleFiles(ShowDirectoriesFirst), failureMessage), qPrintable(failureMessage));
     QQuickFileDialogDelegate *subDirDelegate = nullptr;
     QTRY_VERIFY(findViewDelegateItem(dialogHelper.fileDialogListView, 0, subDirDelegate));
     QVERIFY(doubleClickButton(subDirDelegate));
@@ -1168,15 +1200,17 @@ void tst_QQuickFileDialogImpl::itemsDisabledWhenNecessary()
 void tst_QQuickFileDialogImpl::fileMode_data()
 {
     QTest::addColumn<QQuickFileDialog::FileMode>("fileMode");
+    QTest::addColumn<QString>("acceptButtonText");
 
-    QTest::newRow("OpenFile") << QQuickFileDialog::OpenFile;
-    QTest::newRow("OpenFiles") << QQuickFileDialog::OpenFiles;
-    QTest::newRow("SaveFile") << QQuickFileDialog::SaveFile;
+    QTest::newRow("OpenFile") << QQuickFileDialog::OpenFile << "Open";
+    QTest::newRow("OpenFiles") << QQuickFileDialog::OpenFiles << "Open";
+    QTest::newRow("SaveFile") << QQuickFileDialog::SaveFile << "Save";
 }
 
 void tst_QQuickFileDialogImpl::fileMode()
 {
     QFETCH(QQuickFileDialog::FileMode, fileMode);
+    QFETCH(QString, acceptButtonText);
 
     // Open the dialog.
     FileDialogTestHelper dialogHelper(this, "fileDialog.qml");
@@ -1191,13 +1225,13 @@ void tst_QQuickFileDialogImpl::fileMode()
     COMPARE_URL(dialogHelper.dialog->currentFile(), QUrl::fromLocalFile(tempFile1->fileName()));
     COMPARE_URLS(dialogHelper.dialog->currentFiles(), { QUrl::fromLocalFile(tempFile1->fileName()) });
 
-    // All modes should support opening an existing file, so the Open button should be enabled.
+    // All modes should support opening an existing file, so the accept button should be enabled.
     QVERIFY(dialogHelper.quickDialog->footer());
     auto dialogButtonBox = dialogHelper.quickDialog->footer()->findChild<QQuickDialogButtonBox*>();
     QVERIFY(dialogButtonBox);
-    QQuickAbstractButton* openButton = findDialogButton(dialogButtonBox, "Open");
-    QVERIFY(openButton);
-    QCOMPARE(openButton->isEnabled(), true);
+    QQuickAbstractButton *acceptButton = findDialogButton(dialogButtonBox, acceptButtonText);
+    QVERIFY(acceptButton);
+    QCOMPARE(acceptButton->isEnabled(), true);
 
     // Only the OpenFiles mode should allow multiple files to be selected, however.
     QQuickFileDialogDelegate *tempFile2Delegate = nullptr;
@@ -1342,6 +1376,9 @@ void tst_QQuickFileDialogImpl::setSelectedFile()
     };
     FileDialogTestHelper dialogHelper(
         this, "setSelectedFile.qml", {}, initialProperties);
+
+    dialogHelper.dialog->setOptions(QFileDialogOptions::DontConfirmOverwrite);
+
     OPEN_QUICK_DIALOG();
 
     // The selected file should be what we set.
@@ -1396,6 +1433,7 @@ void tst_QQuickFileDialogImpl::selectNewFileViaTextField_data()
 void tst_QQuickFileDialogImpl::selectNewFileViaTextField()
 {
     QFETCH(QQuickFileDialog::FileMode, fileMode);
+    QFETCH(QString, acceptButtonText);
 
     // Open the dialog.
     FileDialogTestHelper dialogHelper(this, "fileDialog.qml");
@@ -1415,22 +1453,167 @@ void tst_QQuickFileDialogImpl::selectNewFileViaTextField()
              "The TextField for file name should only be visible when the FileMode is 'SaveFile'");
 
     if (fileMode == QQuickFileDialog::SaveFile) {
+        QVERIFY(dialogHelper.quickDialog->footer());
+        auto dialogButtonBox = dialogHelper.quickDialog->footer()->findChild<QQuickDialogButtonBox*>();
+        QVERIFY(dialogButtonBox);
+        QQuickAbstractButton *acceptButton = findDialogButton(dialogButtonBox, acceptButtonText);
+        QVERIFY(acceptButton);
+        QCOMPARE(acceptButton->isEnabled(), false);
+
         const QPoint textFieldCenterPos =
                 fileNameTextField->mapToScene({ fileNameTextField->width() / 2, fileNameTextField->height() / 2 }).toPoint();
 
         QTest::mouseClick(dialogHelper.window(), Qt::LeftButton, Qt::NoModifier, textFieldCenterPos);
         QTRY_VERIFY(fileNameTextField->hasActiveFocus());
+        QCOMPARE(acceptButton->isEnabled(), false);
 
         const QByteArray newFileName("foo.txt");
         for (const auto &c : newFileName)
             QTest::keyClick(dialogHelper.window(), c);
+        QCOMPARE(acceptButton->isEnabled(), true);
+
         QTest::keyClick(dialogHelper.window(), Qt::Key_Enter);
+        QCOMPARE(acceptButton->isEnabled(), true);
 
         QTRY_COMPARE(fileNameTextField->text(), newFileName);
         QCOMPARE(dialogHelper.dialog->selectedFile().fileName(), newFileName);
+
+        QVERIFY(fileNameTextField->hasActiveFocus());
+        for (int i = 0; i < newFileName.size(); i++)
+            QTest::keyClick(dialogHelper.window(), Qt::Key_Backspace);
+        QCOMPARE(acceptButton->isEnabled(), false);
     }
 }
 
+void tst_QQuickFileDialogImpl::selectExistingFileShouldWarnUserWhenFileModeEqualsSaveFile()
+{
+    FileDialogTestHelper dialogHelper(this, "fileDialog.qml");
+    dialogHelper.dialog->setFileMode(QQuickFileDialog::SaveFile);
+    dialogHelper.dialog->setSelectedFile(QUrl::fromLocalFile(tempFile1->fileName()));
+
+    OPEN_QUICK_DIALOG();
+    QQuickTest::qWaitForPolish(dialogHelper.window());
+
+    QSignalSpy acceptedSpy(dialogHelper.dialog, SIGNAL(accepted()));
+
+    auto *dialogButtonBox = dialogHelper.quickDialog->footer()->findChild<QQuickDialogButtonBox *>();
+    QVERIFY(dialogButtonBox);
+
+    auto *confirmationDialog = dialogHelper.quickDialog->findChild<QQuickDialog *>("confirmationDialog");
+    QVERIFY(confirmationDialog);
+
+    auto *openButton = dialogButtonBox->standardButton(QPlatformDialogHelper::Open);
+    QVERIFY(openButton);
+
+    auto *confirmationButtonBox = qobject_cast<QQuickDialogButtonBox *>(confirmationDialog->footer());
+    QVERIFY(confirmationButtonBox);
+
+    const QPoint openButtonCenterPos =
+            openButton->mapToScene({ openButton->width() / 2, openButton->height() / 2 }).toPoint();
+
+    QTest::mouseClick(dialogHelper.window(), Qt::LeftButton, Qt::NoModifier, openButtonCenterPos);
+
+    QTRY_VERIFY(confirmationDialog->isOpened());
+    QVERIFY(dialogHelper.dialog->isVisible());
+
+    // Yes button should have focus by default
+    QTest::keyClick(dialogHelper.window(), Qt::Key_Space, Qt::NoModifier);
+
+    QTRY_VERIFY(!confirmationDialog->isOpened());
+    QVERIFY(!dialogHelper.dialog->isVisible());
+    QCOMPARE(acceptedSpy.count(), 1);
+
+    // Try again, but click "No" this time.
+    QVERIFY(dialogHelper.openDialog());
+    QTRY_VERIFY(dialogHelper.isQuickDialogOpen());
+
+    QTest::keyClick(dialogHelper.window(), Qt::Key_Enter, Qt::NoModifier);
+
+    QTRY_VERIFY(confirmationDialog->isOpened());
+    QVERIFY(dialogHelper.dialog->isVisible());
+
+    // Make No button trigger a clicked() event.
+    auto *confirmationNoButton = confirmationButtonBox->standardButton(QPlatformDialogHelper::No);
+    QVERIFY(confirmationNoButton);
+    QVERIFY(clickButton(confirmationNoButton));
+
+    // FileDialog is still opened
+    QTRY_VERIFY(!confirmationDialog->isOpened());
+    QVERIFY(dialogHelper.dialog->isVisible());
+    QCOMPARE(acceptedSpy.count(), 1);
+
+    // Try again
+    QTest::keyClick(dialogHelper.window(), Qt::Key_Enter, Qt::NoModifier);
+
+    QTRY_VERIFY(confirmationDialog->isOpened());
+    QVERIFY(dialogHelper.dialog->isVisible());
+
+    QTest::keyClick(dialogHelper.window(), Qt::Key_Space, Qt::NoModifier);
+
+    QTRY_VERIFY(!confirmationDialog->isOpened());
+    QVERIFY(!dialogHelper.dialog->isVisible());
+    QCOMPARE(acceptedSpy.count(), 2);
+
+    // Make sure that DontConfirmOverwrite works
+    dialogHelper.dialog->setOptions(QFileDialogOptions::DontConfirmOverwrite);
+
+    QVERIFY(dialogHelper.openDialog());
+    QTRY_VERIFY(dialogHelper.isQuickDialogOpen());
+
+    QTest::keyClick(dialogHelper.window(), Qt::Key_Enter, Qt::NoModifier);
+    QTRY_VERIFY(!confirmationDialog->isOpened());
+    QVERIFY(!dialogHelper.dialog->isVisible());
+    QCOMPARE(acceptedSpy.count(), 3);
+}
+
+void tst_QQuickFileDialogImpl::fileNameTextFieldOnlyChangesWhenSelectingFiles()
+{
+    const auto tempSubFile1Url = QUrl::fromLocalFile(tempSubFile1->fileName());
+    const auto tempSubDirUrl = QUrl::fromLocalFile(tempSubDir.path());
+    const auto tempFile11Url = QUrl::fromLocalFile(tempFile1->fileName());
+
+    const QVariantMap initialProperties = {
+        { "tempFile1Url", QVariant::fromValue(tempSubFile1Url) },
+        { "fileMode", QVariant::fromValue(QQuickFileDialog::SaveFile) }
+    };
+    FileDialogTestHelper dialogHelper(this, "setSelectedFile.qml", {}, initialProperties);
+
+    OPEN_QUICK_DIALOG();
+    QQuickTest::qWaitForPolish(dialogHelper.window());
+
+    QQuickTextField *fileNameTextField =
+            dialogHelper.quickDialog->findChild<QQuickTextField *>("fileNameTextField");
+    QVERIFY(fileNameTextField);
+
+    auto getSelectedFileInfo = [&dialogHelper]() {
+        return QFileInfo(dialogHelper.dialog->selectedFile().toLocalFile());
+    };
+
+    QVERIFY(getSelectedFileInfo().isFile());
+    QCOMPARE(fileNameTextField->text(), tempSubFile1Url.fileName());
+    QCOMPARE(dialogHelper.dialog->selectedFile(), tempSubFile1Url);
+
+    auto *breadcrumbBar = dialogHelper.quickDialog->findChild<QQuickFolderBreadcrumbBar *>();
+    QVERIFY(breadcrumbBar);
+
+    // Pressing the up button causes tempSubDir to be selected
+    QVERIFY(clickButton(breadcrumbBar->upButton()));
+
+    QVERIFY(getSelectedFileInfo().isDir());
+    QCOMPARE(fileNameTextField->text(), tempSubFile1Url.fileName());
+    QCOMPARE(dialogHelper.dialog->selectedFile(), tempSubDirUrl);
+
+    // Change the selected file from the outside
+    dialogHelper.dialog->close();
+    dialogHelper.dialog->setSelectedFile(tempFile11Url);
+    dialogHelper.openDialog();
+    QTRY_VERIFY(dialogHelper.isQuickDialogOpen());
+
+    QVERIFY(getSelectedFileInfo().isFile());
+    QCOMPARE(fileNameTextField->text(), tempFile11Url.fileName());
+    QCOMPARE(dialogHelper.dialog->selectedFile(), tempFile11Url);
+}
+
 QTEST_MAIN(tst_QQuickFileDialogImpl)
 
 #include "tst_qquickfiledialogimpl.moc"
